import React, {Component} from 'react'
import PropTypes from 'prop-types'
import { is, fromJS } from 'immutable'
import { notification, Modal } from 'antd'

import Api from '@/api'
import Utils from '@/utils/utils.js'
import UtilsDM from '@/utils/utils-datamanage.js'
import asyncComponent from '@/utils/asyncComponent'
import MKEmitter from '@/utils/events.js'
import MainTable from './normalTable'
import './index.scss'

// 通用组件
const MainSearch = asyncComponent(() => import('@/tabviews/zshare/topSearch'))
const MainAction = asyncComponent(() => import('@/tabviews/zshare/actionList'))
const NormalHeader = asyncComponent(() => import('@/tabviews/custom/components/share/normalheader'))

class EditableTable extends Component {
  static propTpyes = {
    config: PropTypes.object,        // 组件配置信息
  }

  state = {
    BID: '',              // 上级ID
    BData: '',            // 上级组件行数据
    config: {},           // 页面配置信息，包括按钮、搜索、显示列、标签等
    actions: null,        // 按钮集
    columns: null,        // 显示列
    setting: null,        // 页面全局设置：数据源、按钮及显示列固定、主键等
    data: [],             // 列表数据集
    selectedData: [],     // 已选表格数据
    total: 0,             // 总数
    loading: false,       // 列表数据加载中
    pageIndex: 1,         // 页码
    pageSize: 10,         // 每页数据条数
    orderBy: '',          // 排序
    search: '',           // 搜索条件数组，使用时需分场景处理
    allSearch: null
  }

  /**
   * @description 初始化处理
   * 1、 initdata 为打印时使用的数据集
   */
  UNSAFE_componentWillMount () {
    let _config = fromJS(this.props.config).toJS()
    let setting = {..._config.setting, ..._config.wrap}
    setting.tableId = Utils.getuuid()

    let BID = ''
    let BData = ''

    if (_config.setting.supModule) {
      BData = window.GLOB.CacheData.get(_config.setting.supModule)
    } else {
      BData = window.GLOB.CacheData.get(_config.$pageId)
    }
    if (BData) {
      BID = BData.$BID || ''
    }

    _config.submit.style = _config.submit.style || {}
    _config.submit.wrapStyle = {}
    setting.hasAction = _config.action.length > 0

    _config.action.forEach(btn => {
      if (btn.funcType === 'addline') {
        setting.addable = true
      }
      btn.$tableId = setting.tableId
      if (!btn.controlField) {
        btn.controlField = '$lock'
        btn.controlVals = []
      }
    })

    let _columns = []
    setting.initId = ''
    let triMap = new Map()
    setting.hasSubmit = false

    let getColumns = (cols) => {
      return cols.filter(item => {
        if (item.Hide === 'true') return false

        item.tableId = setting.tableId

        if (item.type === 'colspan') {
          item.subcols = getColumns(item.subcols)

          if (item.subcols.length === 0) {
            return false
          }
        } else if (item.type === 'custom') {
          item.elements.forEach(btn => {
            if (btn.eleType !== 'button') return
            if (btn.funcType === 'addline') {
              setting.addable = true
            }
            btn.$tableId = setting.tableId
            if (!btn.controlField) {
              btn.controlField = '$lock'
              btn.controlVals = []
            }
          })
        } else {
          if (item.type === 'index') {
            item.field = '$Index'
            item.type = 'text'
          }

          if (item.marks && item.marks.length === 0) {
            item.marks = ''
          }

          if (item.editable === 'true') {
            setting.hasSubmit = true
            if (!setting.initId) {
              setting.initId = item.uuid
            }
            if (item.type === 'text' && (item.editType === 'switch' || item.editType === 'select')) {
              triMap.set(item.uuid, item.editType)
            }
            if (item.type === 'text' && item.editType === 'select') {
              item.options = item.options || []
              item.options = item.options.filter(cell => {
                cell.value = cell.Value
                cell.label = cell.Text
        
                return !cell.Hide
              })
            } else if (item.editType === 'popSelect') {
              if (item.pops) {
                item.pops.forEach(cell => {
                  this.resetPop(cell)
                  cell.tabName = cell.initval.replace(/^tab:/, '')
                })
              } else {
                this.resetPop(item)
              }
            }
          }
        }
  
        return true
      })
    }

    _columns = getColumns(_config.cols)

    if (triMap.size > 0) {
      let setColumns = (cols) => {
        return cols.map(item => {
          if (item.type === 'colspan') {
            item.subcols = setColumns(item.subcols)
          } else if (item.editable === 'true' && triMap.has(item.enter)) {
            item.triType = 'click'
          }
    
          return item
        })
      }
  
      _columns = setColumns(_columns)

      if (setting.initId && triMap.has(setting.initId)) {
        setting.triType = 'click'
      }
    }

    if (!_config.lineMarks || _config.lineMarks.length === 0) {
      _config.lineMarks = null
    }

    let allSearch = null
    if (_config.colsCtrls && _config.colsCtrls.length > 0) {
      allSearch = [{key: 'BID', value: BID || ''}]
      allSearch.push(..._config.$searches)
      if (_config.setting.useMSearch) {
        let mainSearch = window.GLOB.SearchBox.get(_config.$searchId) || []
        allSearch.push(...mainSearch)
      }
    } else {
      _config.colsCtrls = null
    }

    if (setting.commit === 'blur') {
      setting.hasSubmit = false
    }

    this.setState({
      pageSize: setting.pageSize || 10,
      BID: BID || '',
      BData: BData || '',
      title: _config.wrap.title,
      config: _config,
      setting: setting,
      actions: _config.action,
      columns: _columns,
      search: _config.$searches,
      allSearch
    }, () => {
      if (_config.setting.onload === 'true') {
        setTimeout(() => {
          this.loadmaindata()
        }, _config.setting.delay || 0)
      }
    })
  }

  resetPop = (config) => {
    let arrfield = config.columns.map(f => f.field)
  
    if (config.linkSubField && config.linkSubField.length > 0) {
      config.linkSubField.forEach(n => {
        if (!arrfield.includes(n)) {
          arrfield.push(n)
        }
      })
    }

    if (sessionStorage.getItem('dataM') === 'true') { // 数据权限
      config.dataSource = config.dataSource.replace(/\$@/ig, '/*').replace(/@\$/ig, '*/').replace(/@datam@/ig, '\'Y\'')
    } else {
      config.dataSource = config.dataSource.replace(/@\$|\$@/ig, '').replace(/@datam@/ig, '\'\'')
    }

    config.dataSource = config.dataSource.replace(/@LoginUID@/ig, `'${sessionStorage.getItem('LoginUID') || ''}'`)
    config.dataSource = config.dataSource.replace(/@SessionUid@/ig, `'${localStorage.getItem('SessionUid') || ''}'`)
    config.dataSource = config.dataSource.replace(/@UserID@/ig, `'${sessionStorage.getItem('UserID') || ''}'`)
    config.dataSource = config.dataSource.replace(/@Appkey@/ig, `'${window.GLOB.appkey || ''}'`)

    if (/\s/.test(config.dataSource)) { // 拼接别名
      config.dataSource = '(' + config.dataSource + ') tb'
    }

    let labels = {}
    config.cols = []
    config.columns.forEach(col => {
      labels[col.field] = col.label

      if (col.Hide === 'true') return

      config.cols.push({
        dataIndex: col.field,
        title: col.label,
        sorter: col.IsSort === 'true',
        width: col.Width || 120
      })
    })

    let placeholder = ''
    if (!config.searchKey) {
      config.onload = 'true'
    } else {
      placeholder = []
      config.searchKey.split(',').forEach(key => {
        if (!labels[key]) {
          placeholder = ''
        } else if (placeholder) {
          placeholder.push(labels[key])
        }
      })

      placeholder = placeholder ? placeholder.join('、') : ''
    }

    config.placeholder = placeholder
    config.arr_field = arrfield.join(',')
  }

  /**
   * @description 主表数据加载
   * @param { Boolean } reset  表格是否重置
   * @param { String }  repage 表格是否重置页码
   */
  async loadmaindata (reset, repage) {
    const { setting, config, search, orderBy, BID, pageIndex, pageSize, BData } = this.state

    if (setting.supModule && !BID) { // BID 不存在时，不做查询
      this.setState({
        data: [],
        selectedData: [],
        total: 0
      })
      reset && MKEmitter.emit('resetTable', config.uuid, 'true') // 列表重置
      return
    }

    let searches = fromJS(search).toJS()
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    if (config.$s_req && searches.filter(item => item.required && item.value === '').length > 0) {
      return
    }

    this.setState({
      loading: true
    })

    let _orderBy = orderBy || setting.order
    let param = UtilsDM.getQueryDataParams(setting, searches, _orderBy, pageIndex, pageSize, BID)

    let result = await Api.genericInterface(param)
    if (result.status) {
      reset && MKEmitter.emit('resetTable', config.uuid, repage) // 列表重置

      let start = 1
      if (setting.laypage) {
        start = pageSize * (pageIndex - 1) + 1
      }
      let data = result.data.map((item, index) => {
        item.key = index
        item.$$uuid = item[setting.primaryKey] || ''
        item.$$BID = BID || ''
        item.$$BData = BData || ''
        item.$Index = start + index + ''
        item.$type = 'upt'
        item.$origin = true

        if (config.absFields) {
          config.absFields.forEach(f => {
            if (!isNaN(item[f])) {
              item[f] = Math.abs(item[f])
            }
          })
        }

        return item
      })

      let total = result.total || 0
      if (config.setting.custompage && data.length) {
        total = data[data.length - 1].mk_total || 0
      }

      let allSearch = null
      if (config.colsCtrls) {
        allSearch = [{key: 'BID', value: BID || ''}]
        allSearch.push(...searches)
      }

      this.setState({
        data: data,
        selectedData: [],
        allSearch: allSearch,
        total: total,
        loading: false
      })

      MKEmitter.emit('transferData', config.uuid, data)

      if (result.message) {
        if (result.ErrCode === 'Y') {
          Modal.success({
            title: result.message
          })
        } else if (result.ErrCode === 'S') {
          notification.success({
            top: 92,
            message: result.message,
            duration: 2
          })
        }
      }
    } else {
      this.setState({
        loading: false
      })
      
      if (!result.message) return
      if (result.ErrCode === 'N') {
        Modal.error({
          title: result.message,
        })
      } else if (result.ErrCode !== '-2') {
        notification.error({
          top: 92,
          message: result.message,
          duration: 10
        })
      }
    }
  }

  /**
   * @description 获取单行数据
   */ 
  async loadmainLinedata (id) {
    const { setting, config, search, orderBy, BID, pageIndex, pageSize, BData } = this.state

    if (config.forbidLine) {
      this.reloadtable()
      return
    }
    
    let searches = fromJS(search).toJS()
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    this.setState({
      loading: true
    })

    let _orderBy = orderBy || setting.order
    let param = UtilsDM.getQueryDataParams(setting, searches, _orderBy, pageIndex, pageSize, BID, id)

    let result = await Api.genericInterface(param)
    if (result.status) {
      if (!result.data || !result.data[0]) {
        this.setState({
          loading: false
        })

        return
      }

      let data = fromJS(this.state.data).toJS()
      let selectedData = fromJS(this.state.selectedData).toJS()

      let _data = result.data[0]
      _data.$$uuid = _data[setting.primaryKey] || ''
      _data.$$BID = BID || ''
      _data.$$BData = BData || ''
      _data.$type = 'upt'
      _data.$origin = true

      if (config.absFields) {
        config.absFields.forEach(f => {
          if (!isNaN(_data[f])) {
            _data[f] = Math.abs(_data[f])
          }
        })
      }

      data = data.map(item => {
        if (item.$$uuid === _data.$$uuid) {
          _data.key = item.key
          _data.$Index = item.$Index
          return _data
        } else {
          return item
        }
      })
      selectedData = selectedData.map(item => {
        if (_data.$$uuid === item.$$uuid) {
          return _data
        }
        return item
      })

      MKEmitter.emit('transferData', config.uuid, _data, 'line')
      MKEmitter.emit('resetSelectLine', config.uuid, _data.$$uuid, _data)

      this.setState({
        data,
        selectedData,
        loading: false
      })
    } else {
      this.setState({
        loading: false
      })
      notification.error({
        top: 92,
        message: result.message,
        duration: 10
      })
    }
  }

  /**
   * @description 搜索条件改变时，重置表格数据
   * 含有初始不加载的页面，修改设置
   */
  refreshbysearch = (searches) => {
    this.setState({
      pageIndex: 1,
      search: searches
    }, () => {
      this.loadmaindata(true, 'true')
    })
  }

  /**
   * @description 表格条件改变时重置数据（分页或排序）
   */
  refreshbytable = (pagination, filters, sorter) => {
    if (sorter.order) {
      let _chg = {
        ascend: 'asc',
        descend: 'desc'
      }
      sorter.order = _chg[sorter.order]
    }

    this.setState({
      pageIndex: pagination.current,
      pageSize: pagination.pageSize,
      orderBy: (sorter.field && sorter.order) ? `${sorter.field} ${sorter.order}` : ''
    }, () => {
      this.loadmaindata()
    })
  }

  /**
   * @description 表格刷新
   */
  reloadtable = (btn) => {
    if (!btn || btn.resetPageIndex !== 'false') {
      this.setState({
        pageIndex: 1
      }, () => {
        this.loadmaindata(true, 'true')
      })
    } else {
      this.loadmaindata(true, 'false')
    }
  }

  /**
   * @description 导出Excel时，获取页面搜索排序等参数
   */
  queryModuleParam = (menuId, callback) => {
    const { config, orderBy, search, setting } = this.state

    if (config.uuid !== menuId) return

    let searches = search ? fromJS(search).toJS() : []
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    callback({
      orderBy: orderBy || setting.order,
      search: searches
    })
  }

  reloadData = (menuId, id) => {
    const { config } = this.state

    if (config.uuid !== menuId) return

    if (!id) {
      this.reloadtable()
    } else {
      this.loadmainLinedata(id)
    }
  }

  resetParentParam = (MenuID, id, data) => {
    const { setting } = this.state

    if (!setting.supModule || setting.supModule !== MenuID) return
    if (id !== this.state.BID || id !== '') {
      this.setState({
        pageIndex: 1,
        BID: id,
        BData: data
      }, () => {
        setTimeout(() => {
          this.loadmaindata(true, 'true')
        }, setting.delay || 0)
      })
    }
  }

  /**
   * @description 按钮执行完成后页面刷新
   * @param {*} menuId     // 菜单Id
   * @param {*} position   // 刷新位置
   * @param {*} btn        // 执行的按钮
   */
  refreshByButtonResult = (menuId, position, btn, id, lines) => {
    const { config, BID } = this.state

    if (config.uuid !== menuId) return

    if (position === 'line') {
      if (lines && lines.length === 1) {
        this.loadmainLinedata(lines[0].$$uuid)
      } else {
        this.reloadtable(btn)
      }
    } else if ((position === 'mainline' || position === 'popclose') && config.setting.supModule && BID) { // 刷新源组件时，附带刷新上级行与当前组件
      MKEmitter.emit('reloadData', config.setting.supModule, BID)
    } else {
      this.reloadtable(btn)
    }
  }

  refreshLineData = (menuId, btn, uuid, count) => {
    const { config } = this.state

    if (config.uuid !== menuId) return
    
    let _data = fromJS(this.state.data).toJS().map(item => {
      if (item.$$uuid === uuid) {
        item[btn.field] = count
      }
      return item
    })
    let _selectedData = fromJS(this.state.selectedData).toJS().map(item => {
      if (item.$$uuid === uuid) {
        item[btn.field] = count
      }
      return item
    })

    this.setState({
      data: _data,
      selectedData: _selectedData
    })
  }

  searchRefresh = (searchId) => {
    const { config } = this.state

    if (config.$searchId !== searchId) return
    
    this.setState({pageIndex: 1}, () => {
      this.reloadtable()
    })
  }

  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.props), fromJS(nextProps)) || !is(fromJS(this.state), fromJS(nextState))
  }

  componentDidMount () {
    const { config } = this.state

    MKEmitter.addListener('reloadData', this.reloadData)
    MKEmitter.addListener('resetSelectLine', this.resetParentParam)
    MKEmitter.addListener('queryModuleParam', this.queryModuleParam)
    MKEmitter.addListener('refreshLineData', this.refreshLineData)
    MKEmitter.addListener('refreshByButtonResult', this.refreshByButtonResult)

    if (config.setting.useMSearch) {
      MKEmitter.addListener('searchRefresh', this.searchRefresh)
    }
  }

  /**
   * @description 组件销毁，清除state更新，清除快捷键设置
   */
  componentWillUnmount () {
    this.setState = () => {
      return
    }
    MKEmitter.removeListener('reloadData', this.reloadData)
    MKEmitter.removeListener('searchRefresh', this.searchRefresh)
    MKEmitter.removeListener('refreshLineData', this.refreshLineData)
    MKEmitter.removeListener('resetSelectLine', this.resetParentParam)
    MKEmitter.removeListener('queryModuleParam', this.queryModuleParam)
    MKEmitter.removeListener('refreshByButtonResult', this.refreshByButtonResult)
  }

  render() {
    const { BID, setting, actions, config, columns, BData, data, selectedData, allSearch } = this.state

    let style = {...config.style}
    if (config.wrap.empty === 'hidden' && data.length === 0) {
      style.display = 'none'
    }

    return (
      <div className={'custom-edit-table' + (setting.hasSubmit ? '' : ' withnot-submit')} id={'anchor' + config.uuid} style={style}>
        <NormalHeader config={config}/>
        {config.search && config.search.length ?
          <MainSearch BID={BID} config={config} refreshdata={this.refreshbysearch}/> : null
        }
        {actions.length > 0 ? <MainAction
          BID={BID}
          setting={setting}
          actions={actions}
          BData={BData}
          columns={config.columns}
          selectedData={selectedData}
        /> : null}
        <MainTable
          BID={BID}
          setting={setting}
          columns={columns}
          MenuID={config.uuid}
          allSearch={allSearch}
          submit={config.submit}
          fields={config.columns}
          total={this.state.total}
          colsCtrls={config.colsCtrls}
          lineMarks={config.lineMarks}
          loading={this.state.loading}
          refreshdata={this.refreshbytable}
          chgSelectData={(selects) => this.setState({selectedData: selects})}
        />
      </div>
    )
  }
}

export default EditableTable